# coding: utf-8
#
# This code is part of pyrekordbox.
#
# Copyright (c) 2022, Dylan Jones

"""Rekordbox XML database file interface."""

import logging
import os.path
import urllib.parse
from collections import abc
from xml.dom import minidom
import xml.etree.cElementTree as xml
import bidict

logger = logging.getLogger(__name__)

URL_PREFIX = "file://localhost/"

POSMARK_TYPE_MAPPING = bidict.bidict(
    {
        "0": "cue",
        "1": "fadein",
        "2": "fadeout",
        "3": "load",
        "4": "loop",
    }
)


RATING_MAPPING = bidict.bidict(
    {"0": 0, "51": 1, "102": 2, "153": 3, "204": 4, "255": 5}
)


NODE_KEYTYPE_MAPPING = bidict.bidict({"0": "TrackID", "1": "Location"})


def pretty_xml(element, indent=None, encoding="utf-8"):
    """Generates a formatted string of an XML element.

    Parameters
    ----------
    element : xml.etree.cElementTree.Element
        The input XML element.
    indent : str, optional
        The indentation used for formatting the XML element. The default is 3 spaces.
    encoding : str, optional
        The encoding used for formatting the XML element. The default is `utf-8`.

    Returns
    -------
    xml_string : str
        The formatted string of the XML element.
    """
    # Build pretty xml-string
    if indent is None:
        indent = "    "
    if encoding is None:
        encoding = "utf-8"
    rough_string = xml.tostring(element, encoding)
    reparsed = minidom.parseString(rough_string)
    string = reparsed.toprettyxml(indent=indent, encoding=encoding).decode()
    # Remove annoying empty lines
    string = "\n".join([line for line in string.splitlines() if line.strip()])
    return string


def encode_path(path):
    """Encodes a file path as URI string.

    Parameters
    ----------
    path : str
        The file path to encode.

    Returns
    -------
    url : str
        The encoded file path as URI string.

    Examples
    --------
    >>> s = r"C:\Music\PioneerDJ\Demo Tracks\Demo Track 1.mp3"  # noqa: W605
    >>> encode_path(s)
    file://localhost/C:/Music/PioneerDJ/Demo%20Tracks/Demo%20Track%201.mp3

    """
    url_path = urllib.parse.quote(path, safe=":/\\")
    url = URL_PREFIX + url_path.replace("\\", "/")
    return url


def decode_path(url):
    """Decodes an as URI string encoded file path.

    Parameters
    ----------
    url : str
        The encoded file path to decode.

    Returns
    -------
    path : str
        The decoded file path.

    Examples
    --------
    >>> s = r"file://localhost/C:/Music/PioneerDJ/Demo%20Tracks/Demo%20Track%201.mp3"
    >>> decode_path(s)
    C:\Music\PioneerDJ\Demo Tracks\Demo Track 1.mp3  # noqa: W605

    """
    path = urllib.parse.unquote(url)
    path = path.replace(URL_PREFIX, "")
    return os.path.normpath(path)


class XmlAttributeKeyError(Exception):
    def __init__(self, cls, key, attributes):
        super().__init__(
            f"{key} is not a valid key for {cls.__name__}! Valid attribs:\n{attributes}"
        )


class AbstractElement(abc.Mapping):
    """Abstract base class for Rekordbox XML elements."""

    TAG: str
    ATTRIBS: list

    SETTERS = dict()
    GETTERS = dict()

    def __init__(self, element=None, *args, **kwargs):
        self._element = element
        if element is None:
            self._init(*args, **kwargs)
        else:
            self._load_subelements()

    def _init(self, *args, **kwargs):
        pass

    def _load_subelements(self):
        pass

    def get(self, key, default=None):
        if key not in self.ATTRIBS:
            raise XmlAttributeKeyError(self.__class__, key, self.ATTRIBS)
        value = self._element.attrib.get(key, default)
        if value == default:
            return default
        try:
            # Apply callback
            value = self.GETTERS[key](value)
        except KeyError:
            pass
        return value

    def set(self, key, value):
        if key not in self.ATTRIBS:
            raise XmlAttributeKeyError(self.__class__, key, self.ATTRIBS)
        try:
            # Apply callback
            value = self.SETTERS[key](value)
        except KeyError:
            # Convert to str just in case
            value = str(value)
        self._element.attrib.set(key, value)

    def __len__(self):
        return len(self._element.attrib)

    def __iter__(self):
        return iter(self._element.attrib.keys())

    def __getitem__(self, key):
        return self.get(key)

    def __setitem__(self, key, value):
        self.set(key, value)

    def __getattr__(self, key):
        return self.get(key)

    def __repr__(self):
        return f"<{self.__class__.__name__}()>"


# -- Collection elements ---------------------------------------------------------------


# noinspection PyPep8Naming
class Tempo(AbstractElement):
    """Tempo element representing the beat grid.

    Attributes
    ----------
    Inizio : float
        The start position of the beat grid item.
    Bpm : float
        The BPM value of the beat grid item.
    Metro : str, optional
        The kind of musical meter, for example '4/4'. The default is '4/4'.
    Battito : int
        The beat number in the bar. If `metro` is '4/4', the value can be 1, 2, 3 or 4.
    """

    TAG = "TEMPO"
    ATTRIBS = ["Inizio", "Bpm", "Metro", "Battito"]
    GETTERS = {"Inizio": float, "Bpm": float, "Battito": int}

    def __init__(
        self, parent=None, Inizio=0.0, Bpm=0.0, Metro="4/4", Battito=1, element=None
    ):
        super().__init__(element, parent, Inizio, Bpm, Metro, Battito)

    def _init(self, parent, inizio, bpm, metro, battito):
        attrib = {
            "Inizio": str(inizio),
            "Bpm": str(bpm),
            "Metro": str(metro),
            "Battito": str(battito),
        }
        self._element = xml.SubElement(parent, self.TAG, attrib=attrib)

    def __repr__(self):
        args = ", ".join(
            [
                f"Inizio={self.Inizio}",
                f"Bpm={self.Bpm}",
                f"Metro={self.Metro}",
                f"Battito={self.Battito}",
            ]
        )
        return f"<{self.__class__.__name__}({args})>"


# noinspection PyPep8Naming
class PositionMark(AbstractElement):
    """Position element for storing position markers like cue points.

    Attributes
    ----------
    Name : str
        The name of the position mark.
    Type : str
        The type of position mark. Can be 'cue', 'fadein', 'fadeout', 'load' or 'loop'.
    Start : float
        Start position of the position mark in seconds.
    End : float, optionl
        End position of the position mark in seconds.
    Num : int, optional
        Charakter for identification of the position mark (for hot cues). For memory
        cues this is always -1.
    """

    TAG = "POSITION_MARK"
    ATTRIBS = ["Name", "Type", "Start", "End", "Num"]

    GETTERS = {
        "Type": POSMARK_TYPE_MAPPING.get,
        "Start": float,
        "End": float,
        "Num": int,
    }
    SETTERS = {"Type": POSMARK_TYPE_MAPPING.inv.get}  # noqa

    def __init__(
        self,
        parent=None,
        Name="",
        Type="cue",
        Start=0.0,
        End=None,
        Num=-1,
        element=None,
    ):
        super().__init__(element, parent, Name, Type, Start, End, Num)

    def _init(self, parent, name, type_, start, end, num):
        attrib = {
            "Name": name,
            "Type": POSMARK_TYPE_MAPPING.inv.get(type_),  # noqa
            "Start": str(start),
            "Num": str(num),
        }
        if end is not None:
            attrib["End"] = str(end)
        self._element = xml.SubElement(parent, self.TAG, attrib=attrib)

    def __repr__(self):
        args = ", ".join(
            [
                f"Name={self.Name}",
                f"Type={self.Type}",
                f"Start={self.Start}",
                f"End={self.End}",
                f"Num={self.Num}",
            ]
        )
        return f"<{self.__class__.__name__}({args})>"


# noinspection PyPep8Naming
class Track(AbstractElement):
    """Track element for storing the metadata of a track.

    Attributes
    ----------
    TrackID : int
        Identification of the track.
    Name: str
        The name of the track.
    Artist : str
        The name of the artist.
    Composer : str
        The name of the composer (or producer).
    Album : str
        The name of the album.
    Grouping : str
        The name of the grouping.
    Genre : str
        The name of the genre.
    Kind : str
        The kind of the audio file, for example 'WAV File' or 'MP3 File'.
    Size : int
        The size of the audio file.
    TotalTime : int
        The duration of the track in seconds.
    DiscNumber : int
        The number of the disc of the album.
    TrackNumber : int
        The Number of the track of the album.
    Year : int
        The year of release.
    AverageBpm : float
        The average BPM of the track.
    DateModified : str
        The date of last modification in the format 'yyyy-mm-dd'.
    DateAdded : str
        The date of addition modification in the format 'yyyy-mm-dd'.
    BitRate : int
        The encoding bit rate.
    SampleRate : float
        The frequency of sampling.
    Comments : str
        The comments of the track.
    PlayCount : int
        The play count of the track.
    LastPlayed : str
        The date of last playing in the format 'yyyy-mm-dd'.
    Rating : int
        The rating of the track using the mapping 0=0, 1=51, 2=102, 3=153, 4=204, 5=255.
    Location : str
        The location of the file encoded as URI string. This value is essential for
        each track.
    Remixer : str
        The name of the remixer.
    Tonality : str
        The tonality or kind of musical key.
    Label : str
        The name of the record label.
    Mix : str
        The name of the mix.
    Colour : str
        The color for track grouping in RGB format.
    tempos : list
        The `Tempo` elements of the track.
    marks : list
        The `PositionMark` elements of the track.
    """

    TAG = "TRACK"
    ATTRIBS = [
        "TrackID",
        "Name",
        "Artist",
        "Composer",
        "Album",
        "Grouping",
        "Genre",
        "Kind",
        "Size",
        "TotalTime",
        "DiscNumber",
        "TrackNumber",
        "Year",
        "AverageBpm",
        "DateModified",
        "DateAdded",
        "BitRate",
        "SampleRate",
        "Comments",
        "PlayCount",
        "LastPlayed",
        "Rating",
        "Location",
        "Remixer",
        "Tonality",
        "Label",
        "Mix",
        "Colour",
    ]

    GETTERS = {
        "TrackID": int,
        "Size": int,
        "TotalTime": int,
        "DiscNumber": int,
        "TrackNumber": int,
        "Year": int,
        "AverageBpm": float,
        "BitRate": int,
        "SampleRate": float,
        "PlayCount": int,
        "Rating": RATING_MAPPING.get,
        "Location": decode_path,
    }

    SETTERS = {"Rating": RATING_MAPPING.inv.get, "Location": encode_path}  # noqa

    def __init__(self, parent=None, Location="", element=None, **kwargs):
        self.tempos = list()
        self.marks = list()
        super().__init__(element, parent, Location, **kwargs)

    def _init(self, parent, Location, **kwargs):
        attrib = {"Location": encode_path(Location)}
        for key, val in kwargs.items():
            if key not in self.ATTRIBS:
                raise KeyError(
                    f"{key} is not a valid key for {self.__class__.__name__}!"
                )
            attrib[key] = str(val)
        self._element = xml.SubElement(parent, self.TAG, attrib=attrib)

    def _load_subelements(self):
        tempo_elements = self._element.findall(f"{Tempo.TAG}")
        if tempo_elements is not None:
            self.tempos = [Tempo(element=el) for el in tempo_elements]
        mark_elements = self._element.findall(f".//{PositionMark.TAG}")
        if mark_elements is not None:
            self.marks = [PositionMark(element=el) for el in mark_elements]

    def add_tempo(self, Inizio=0.0, Bpm=0.0, Metro="4/4", Battito=1):
        return Tempo(self, Inizio, Bpm, Metro, Battito)

    def add_mark(self, Name="", Type="cue", Start=0.0, End=None, Num=-1):
        return PositionMark(self, Name, Type, Start, End, Num)

    def __repr__(self):
        return f"<{self.__class__.__name__}(Location={self.Location})>"


# -- Playlist elements -----------------------------------------------------------------


class Node:
    """Node element used for storing playlist folders and playlists."""

    TAG = "NODE"

    FOLDER = 0
    PLAYLIST = 1

    def __init__(self, parent=None, element=None, **attribs):
        self._parent = parent
        self._element = element
        if element is None:
            self._element = xml.SubElement(parent, self.TAG, attrib=attribs)

    @classmethod
    def folder(cls, parent, name):
        attrib = {"Name": name, "Type": str(cls.FOLDER), "Count": "0"}
        return cls(parent, **attrib)

    @classmethod
    def playlist(cls, parent, name, keytype="TrackID"):
        attrib = {
            "Name": name,
            "Type": str(cls.PLAYLIST),
            "KeyType": NODE_KEYTYPE_MAPPING.inv[keytype],  # noqa
            "Entries": "0",
        }
        return cls(parent, **attrib)

    @property
    def parent(self):
        return self._parent

    @property
    def name(self):
        """str: Name of Node."""
        return self._element.attrib.get("Name")

    @property
    def type(self):
        """int: Type of Node (0=folder or 1=playlist)."""
        return int(self._element.attrib.get("Type"))

    @property
    def count(self):
        return int(self._element.attrib.get("Count", 0))

    @property
    def entries(self):
        return int(self._element.attrib.get("Entries", 0))

    @property
    def key_type(self):
        return NODE_KEYTYPE_MAPPING.get(self._element.attrib.get("KeyType"))

    @property
    def is_folder(self):
        return self.type == self.FOLDER

    @property
    def is_playlist(self):
        return self.type == self.PLAYLIST

    def _update_count(self):
        self._element.attrib["Count"] = str(len(self._element))

    def _update_entries(self):
        self._element.attrib["Entries"] = str(len(self._element))

    def get_node(self, i):
        """Returns the i-th sub-Node of the current node.

        Parameters
        ----------
        i : int
            Index of sub-Node

        Returns
        -------
        subnode : Node
        """
        return Node(self, element=self._element.findall(f"{self.TAG}[{i + 1}]"))

    def get_playlist(self, name):
        """Returns the sub-Node with the given name.

        Parameters
        ----------
        name : str
            Name of the sub-Node

        Returns
        -------
        subnode : Node
        """
        return Node(self, element=self._element.find(f'.//{self.TAG}[@Name="{name}"]'))

    def get_playlists(self):
        return [Node(self, element=el) for el in self._element]

    def add_playlist_folder(self, name):
        """Add a new playlist folder as child to this node.

        Parameters
        ----------
        name : str
            The name of the new playlist folder.

        Returns
        -------
        folder_node : Node
            The newly created playlist folder node.
        """
        if self.is_playlist:
            raise ValueError("Sub-elements can only be added to a folder node!")

        node = Node.folder(self._element, name)
        self._update_count()
        return node

    def add_playlist(self, name, keytype="TrackID"):
        """Add a new playlist as child to this node.

        Parameters
        ----------
        name : str
            The name of the new playlist.
        keytype : {'TrackID', 'Location'} str
            The type of key the playlist uses to store the tracks. Can either be
            'TrackID' or 'Location'.

        Returns
        -------
        playlist_node : Node
            The newly created playlist node.
        """
        if self.is_playlist:
            raise ValueError("Sub-elements can only be added to a folder node!")

        node = Node.playlist(self._element, name, keytype)
        self._update_count()
        return node

    def remove_playlist(self, name):
        item = self.get_playlist(name)
        self._element.remove(item._element)  # noqa
        self._update_count()
        self._update_entries()

    def add_track(self, key):
        el = xml.SubElement(self._element, Track.TAG, attrib={"Key": str(key)})
        self._update_entries()
        return el

    def remove_track(self, key):
        el = self._element.find(f'{Track.TAG}[@Key="{key}"]')
        self._element.remove(el)
        self._update_entries()
        return el

    def get_tracks(self):
        if self.type == self.FOLDER:
            return list()
        elements = self._element.findall(f".//{Track.TAG}")
        items = list()
        for el in elements:
            val = el.attrib["Key"]
            if self.key_type == "TrackID":
                val = int(val)
            items.append(val)
        return items

    def get_track(self, key):
        el = self._element.find(f'{Track.TAG}[@Key="{key}"]')
        val = el.attrib["Key"]
        if self.key_type == "TrackID":
            val = int(val)
        return val

    def treestr(self, lvl=0, indent=4):
        space = indent * lvl * " "
        string = ""
        if self.type == self.PLAYLIST:
            string += space + f"Playlist: {self.name} ({self.entries} Tracks)\n"
        elif self.type == self.FOLDER:
            string += space + f"Folder: {self.name}\n"
            for node in self.get_playlists():
                string += node.treestr(lvl + 1, indent)
        return string

    def __eq__(self, other):
        return self.parent == other.parent and self.name == other.name

    def __repr__(self):
        return f"<{self.__class__.__name__}({self.name})>"


# -- Main XML object -------------------------------------------------------------------


# noinspection PyPep8Naming
class RekordboxXml:
    """Rekordbox XML database object.

    The XML database contains the tracks and playlists in the Rekordbox collection. By
    importing the databse, new tracks and items can be added to the Rekordbox
    collection.
    The Rekordbox XML file can be created in Rekordbox 5/6 in the application menu
    'File>Export collection in xml format'. It can be imported by setting the file path
    in 'File>Preferences>Advanced>rekordbox xml>Imported Libary'.

    If a file path is passed to the constructor of the `RekordboxXml` object, the file
    is opened and parsed. Otherwise, an empty file is created with the given arguments.
    Creating an importable XML file requires a product name, xml database version and
    company name.

    Attributes
    ----------
    path : str, optional
        The file path to

    Examples
    --------
    Open Rekordbox XML file

    >>> file = RekordboxXml(os.path.join(".testdata", "rekordbox 6", "database.xml"))

    Create new XML file

    >>> file = RekordboxXml()

    """

    ROOT_TAG = "DJ_PLAYLISTS"
    PRDT_TAG = "PRODUCT"
    PLST_TAG = "PLAYLISTS"
    COLL_TAG = "COLLECTION"

    def __init__(self, path="", name=None, version=None, company=None):
        self._root = None
        self._product = None
        self._collection = None
        self._playlists = None
        self._root_node = None

        self._last_id = 0
        if path:
            self._parse(path)
        else:
            self._init(name, version, company)

    @property
    def frmt_version(self):
        """str : The version of the Rekordbox XML format."""
        return self._root.attrib["Version"]

    @property
    def product_name(self):
        """str : The product name that will be displayed in the software."""
        return self._product.attrib.get("Name")

    @property
    def product_version(self):
        """str : The product version."""
        return self._product.attrib.get("Version")

    @property
    def product_company(self):
        """str : The company name."""
        return self._product.attrib.get("Company")

    @property
    def num_tracks(self):
        """str : The number of tracks in the collection."""
        return int(self._collection.attrib.get("Entries"))

    @property
    def root_playlist_folder(self):
        return self._root_node

    def _parse(self, path):
        """Parse an existing XML file."""
        tree = xml.parse(path)
        self._root = tree.getroot()
        self._product = self._root.find(self.PRDT_TAG)
        self._collection = self._root.find(self.COLL_TAG)
        self._playlists = self._root.find(self.PLST_TAG)
        self._root_node = Node(element=self._playlists.find(Node.TAG))

    def _init(self, name=None, version=None, company=None, frmt_version=None):
        """Initialize a new XML file."""
        frmt_version = frmt_version or "1.0.0"
        name = name or "pyrekordbox"
        version = version or "0.0.1"
        company = company or ""

        # Initialize root element
        self._root = xml.Element(self.ROOT_TAG, attrib={"Version": frmt_version})
        # Initialize product element
        attrib = {"Name": name, "Version": version, "Company": company}
        self._product = xml.SubElement(self._root, self.PRDT_TAG, attrib=attrib)
        # Initialize collection element
        attrib = {"Entries": "0"}
        self._collection = xml.SubElement(self._root, self.COLL_TAG, attrib=attrib)
        # Initialize playlist element
        self._playlists = xml.SubElement(self._root, self.PLST_TAG)
        self._root_node = Node.folder(self._playlists, "ROOT")

        track_ids = self.get_track_ids()
        if track_ids:
            self._last_id = max(track_ids)

    def get_tracks(self):
        """Returns the tracks in the collection of the XML file.

        Returns
        -------
        tracks : list of Track
            A list of the track objects in the collection.
        """
        elements = self._collection.findall(f".//{Track.TAG}")
        return [Track(element=el) for el in elements]

    def get_track(self, index=None, TrackID=None, Location=None):
        """Get a track in the collection of the XML file.

        Parameters
        ----------
        index : int, optional
            If `index` is given, the track with this index in the collection is
            returned.
        TrackID : int, optional
            If `TrackID` is given, the track with this ID in the collection is
            returned.
        Location : str, optional
            If `Location` is given, the track with this file path in the collection is
            returned.

        Returns
        -------
        track : Track
            The XML track element.

        Examples
        --------
        Get track by index

        >>> file = RekordboxXml("database.xml")
        >>> track = file.get_track(0)

        or by ``TrackID``

        >>> track = file.get_track(TrackID=1)

        """
        if index is None and TrackID is None:
            raise ValueError("Either index or TrackID has to be specified!")

        if TrackID is not None:
            el = self._collection.find(f'.//{Track.TAG}[@TrackID="{TrackID}"]')
        elif Location is not None:
            el = self._collection.find(f'.//{Track.TAG}[@Location="{Location}"]')
        else:
            el = self._collection.find(f".//{Track.TAG}[{index + 1}]")
        return Track(element=el)

    def get_track_ids(self):
        """Returns the `TrackID` of all tracks in the collection of the XML file.

        Returns
        -------
        ids : list of int
            The ID's of all tracks.
        """
        elements = self._collection.findall(f".//{Track.TAG}")
        return [int(el.attrib["TrackID"]) for el in elements]

    def get_playlist(self, *names):
        """Returns a playlist or playlist folder with the given path.

        Parameters
        ----------
        *names : str
            Names in the path. If no names are given the root playlist folder is
            returned.

        Returns
        -------
        node : Node
            The playlist or playlist folder node.

        Examples
        --------
        >>> file = RekordboxXml("database.xml")
        >>> playlist = file.get_playlist("Folder", "Sub Playlist")

        """
        node = self._root_node
        if not names:
            return node
        for name in names:
            node = node.get_playlist(name)
        return node

    def _update_track_count(self):
        """Updates the track count element."""
        num_tracks = len(self._collection.findall(f".//{Track.TAG}"))
        self._collection.attrib["Entries"] = str(num_tracks)

    def add_track(self, location, **kwargs):
        """Add a new track element to the Rekordbox XML collection.

        Parameters
        ----------
        location : str
            The file path of the track.
        kwargs :
            Keyword arguments which are used to fill the track attributes. If no
            argument for ``TrackID`` is given the ID is auto-incremented.

        Returns
        -------
        track : Track
            The newly created XML track element.

        Examples
        --------
        >>> file = RekordboxXml("database.xml")
        >>> _ = file.add_track("path/to/track.wav")
        """
        if "TrackID" not in kwargs:
            kwargs["TrackID"] = self._last_id + 1

        # Check that Location and TrackID are unique
        track_id = kwargs["TrackID"]

        for other in self.get_tracks():
            if track_id == other.TrackID:
                raise ValueError(
                    f"XML database already contains a track with TrackID={track_id}!"
                )
            elif os.path.normpath(location) == other.Location:
                raise ValueError(
                    f"XML database already contains a track with Location={location}!"
                )
        # Create track and add it to the collection
        track = Track(self._collection, location, **kwargs)
        self._last_id = int(track["TrackID"])
        self._update_track_count()
        return track

    def remove_track(self, track):
        """Remove a track element from the Rekordbox XML collection.

        Parameters
        ----------
        track : Track
            The XML track element to remove.

        Examples
        --------
        >>> file = RekordboxXml("database.xml")
        >>> t = file.get_track(0)
        >>> file.remove_track(t)

        """
        self._collection.remove(track._element)  # noqa
        self._update_track_count()

    def add_playlist_folder(self, name):
        """Add a new top-level playlist folder to the XML collection.

        Parameters
        ----------
        name : str
            The name of the new playlist folder.

        Returns
        -------
        folder_node : Node
            The newly created playlist folder node.

        See Also
        --------
        Node.add_playlist_folder
        """
        return self._root_node.add_playlist_folder(name)

    def add_playlist(self, name, keytype="TrackID"):
        """Add a new top-level playlist to the XML collection.

        Parameters
        ----------
        name : str
            The name of the new playlist.
        keytype : {'TrackID', 'Location'} str
            The type of key the playlist uses to store the tracks. Can either be
            'TrackID' or 'Location'.

        Returns
        -------
        playlist_node : Node
            The newly created playlist node.

        See Also
        --------
        Node.add_playlist
        """
        return self._root_node.add_playlist(name, keytype)

    def tostring(self, indent=None):
        """Returns the contents of the XML file as a string.

        Parameters
        ----------
        indent : str, optional
            The indentation used for formatting the XML file. The default is 3 spaces.

        Returns
        -------
        s : str
            The contents of the XML file
        """
        return pretty_xml(self._root, indent, encoding="utf-8")

    def save(self, path="", indent=None):
        """Writes the contents of the XML file to disk.

        Parameters
        ----------
        path : str, optional
            The path for saving the XML file. The default is the original file.
        indent : str, optional
            The indentation used for formatting the XML element.
            The default is 3 spaces.
        """
        string = self.tostring(indent)
        with open(path, "w") as fh:
            fh.write(string)

    def __repr__(self):
        name = self.product_name
        v = self.product_version
        company = self.product_company
        tracks = self.num_tracks

        cls = self.__class__.__name__
        s = f"{cls}(tracks={tracks}, info={name}, {company}, v{v})"
        return s



# coding: utf-8
#
# This code is part of pyrekordbox.
#
# Copyright (c) 2022, Dylan Jones

"""Configuration handling for pyrekordbox."""

import os
import logging
from .utils import (
    read_pyrekordbox_configuration,
    get_pioneer_app_dir,
    get_pioneer_install_dir,
    read_rekordbox_settings,
    read_rekordbox6_options,
)

logger = logging.getLogger(__name__)

# Define empty pyrekordbox configuration
__config__ = {
    "pioneer": {
        "app_dir": "",
        "install_dir": "",
    },
    "rekordbox5": {
        "version": "",
        "db_path": "",
        "db_dir": "",
        "app_dir": "",
        "install_dir": "",
    },
    "rekordbox6": {
        "version": "",
        "db_path": "",
        "db_dir": "",
        "app_dir": "",
        "install_dir": "",
        "dp": "",
    },
}


def _get_rb_config(pioneer_prog_dir: str, pioneer_app_dir: str, version: int):
    # Get latest Rekordbox installation directory for major release `version`
    # -----------------------------------------------------------------------

    # Find all 'V.x.x' version strings in dir names
    versions = list()
    for name in os.listdir(pioneer_prog_dir):
        if name.startswith("rekordbox"):
            ver_str = name.replace("rekordbox", "").strip()
            if ver_str.startswith(str(version)):
                versions.append(ver_str)

    # Get latest 'V.x.x' version string and assure there is one
    versions.sort(key=lambda s: list(map(int, s.split("."))))
    try:
        rb_version = versions[-1]
    except IndexError:
        raise FileNotFoundError(
            f"No Rekordbox {version} folder found in {pioneer_prog_dir}"
        )

    # Find Rekordbox installation directory path
    rb_prog_dir = os.path.join(pioneer_prog_dir, f"rekordbox {rb_version}")
    # Check installation directory
    if not os.path.exists(rb_prog_dir):
        raise FileNotFoundError(f"The directory '{rb_prog_dir}' doesn't exist!")

    logger.debug("Found Rekordbox %s install-dir: %s", version, rb_prog_dir)

    # Get Rekordbox application directory path for major release `version`
    # -------------------------------------------------------------------

    name = "rekordbox6" if version == 6 else "rekordbox"
    rb_app_dir = os.path.join(pioneer_app_dir, name)
    if not os.path.exists(rb_app_dir):
        raise FileNotFoundError(f"The directory '{rb_app_dir}' doesn't exist!")

    logger.debug("Found Rekordbox %s app-dir: %s", version, rb_app_dir)

    # Get Rekordbox database locations for major release `version`
    # ------------------------------------------------------------

    settings = read_rekordbox_settings(rb_app_dir)

    db_dir = os.path.normpath(settings["masterDbDirectory"])
    db_filename = "master.db" if version == 6 else "datafile.edb"
    db_path = os.path.join(db_dir, db_filename)
    if not os.path.exists(db_path):
        raise FileNotFoundError(f"The Rekordbox database '{db_path}' doesn't exist!")

    conf = {
        "version": rb_version,
        "install_dir": rb_prog_dir,
        "app_dir": rb_app_dir,
        "db_dir": db_dir,
    }
    return conf


def _get_rb5_config(pioneer_prog_dir: str, pioneer_app_dir: str):
    conf = _get_rb_config(pioneer_prog_dir, pioneer_app_dir, version=5)

    # RB5 database is called 'datafile.edb'
    db_path = os.path.join(conf["db_dir"], "datafile.edb")
    if not os.path.exists(db_path):
        raise FileNotFoundError(f"The Rekordbox database '{db_path}' doesn't exist!")

    conf.update({"db_path": db_path})
    return conf


def _get_rb6_config(pioneer_prog_dir: str, pioneer_app_dir: str):
    conf = _get_rb_config(pioneer_prog_dir, pioneer_app_dir, version=6)

    # Read Rekordbox 6 'options.json'
    opts = read_rekordbox6_options(pioneer_app_dir)
    db_path = os.path.normpath(opts["db-path"])
    db_dir = os.path.dirname(db_path)
    assert conf["db_dir"] == db_dir

    if not os.path.exists(db_path):
        raise FileNotFoundError(f"The Rekordbox database '{db_path}' doesn't exist!")

    conf.update({"db_path": db_path, "dp": opts["dp"]})
    return conf


def update_config(pioneer_install_dir="", pioneer_app_dir=""):
    """Update the pyrekordbox configuration.

    This method scans the system for the Rekordbox installation and application data
    directories and extracts the reuired file locations. For this the default Pioneer
    directories (installation and application data) are used. If the method fails to
    find the directories they can be supplied as parameters. If no Rekordbox
    installation is found the fileds are left unchanged.
    On import configuration with the default locations is loaded.

    Parameters
    ----------
    pioneer_install_dir : str, optional
        The path to the Pioneer installation directory. This is where the program files
        of Pioneer applications are stored. By default, the normal location of Pioneer
        programs is used.
    pioneer_app_dir : str, optional
        The path to the Pioneer application directory. This is where the application
        user data of Pioneer programs is stored. By default, the normal location of
        the Pioneer application data is used.
    """
    conf = read_pyrekordbox_configuration()
    if not pioneer_install_dir and "pioneer-install-dir" in conf:
        pioneer_install_dir = conf["pioneer-install-dir"]
    if not pioneer_app_dir and "pioneer-app-dir" in conf:
        pioneer_app_dir = conf["pioneer-app-dir"]

    try:
        # Pioneer installation directory
        pioneer_install_dir = get_pioneer_install_dir(pioneer_install_dir)
        __config__["pioneer"]["install_dir"] = pioneer_install_dir
    except FileNotFoundError as e:
        logger.warning(e)
        return

    try:
        # Pioneer application data directory
        pioneer_app_dir = get_pioneer_app_dir(pioneer_app_dir)
        __config__["pioneer"]["app_dir"] = pioneer_app_dir
    except FileNotFoundError as e:
        logger.warning(e)
        return

    # Update Rekordbox 5 config
    try:
        conf = _get_rb5_config(pioneer_install_dir, pioneer_app_dir)
        __config__["rekordbox5"].update(conf)
    except FileNotFoundError as e:
        logging.warning(e)

    # Update Rekordbox 6 config
    try:
        conf = _get_rb6_config(pioneer_install_dir, pioneer_app_dir)
        __config__["rekordbox6"].update(conf)
    except FileNotFoundError as e:
        logging.warning(e)


# Fill the pyrekordbox-configuration
update_config()


def get_config(section, key=None):
    """Gets a section or value of the pyrekordbox configuration.

    Parameters
    ----------
    section : str
        The name of the section.
    key : str, optional
        The name of the specific value to return. If not given all values of the section
        are returned as dictionary.

    Returns
    -------
    data : str or dict
        The data of a section or a specific configuration value.
    """
    conf = __config__[section]
    if key is None:
        return conf
    return conf[key]


def pformat_config(indent="   ", hw=14, delim=" = "):
    """Returns a formatted string of the pyrekordbox configurations."""
    pioneer = __config__["pioneer"]
    rb5 = __config__["rekordbox5"]
    rb6 = __config__["rekordbox6"]

    lines = ["Pioneer:"]
    lines += [f"{indent}{k + delim:<{hw}} {pioneer[k]}" for k in sorted(pioneer.keys())]
    lines.append("Rekordbox 5:")
    lines += [f"{indent}{k + delim:<{hw}} {rb5[k]}" for k in sorted(rb5.keys())]
    lines.append("Rekordbox 6:")
    lines += [f"{indent}{k + delim:<{hw}} {rb6[k]}" for k in sorted(rb6.keys())]
    return "\n".join(lines)


def show_config():
    """Prints a formatted string of the pyrekordbox configurations."""
    print(pformat_config())



# coding: utf-8
#
# This code is part of pyrekordbox.
#
# Copyright (c) 2022, Dylan Jones

import os
from pyrekordbox import anlz

TEST_ROOT = os.path.join(os.path.dirname(os.path.dirname(__file__)), ".testdata")
ANLZ_ROOT = os.path.join(TEST_ROOT, "exports", "demo_tracks", "PIONEER", "USBANLZ")

ANLZ_ROOT1 = os.path.join(ANLZ_ROOT, "P016", "0000875E")
ANLZ_ROOT2 = os.path.join(ANLZ_ROOT, "P053", "0001D21F")


def test_parse():
    for root, files in anlz.walk_anlz_paths(ANLZ_ROOT):
        for path in files.values():
            anlz.AnlzFile.parse_file(path)


def test_rebuild():
    for root, files in anlz.walk_anlz_paths(ANLZ_ROOT):
        for path in files.values():
            file = anlz.AnlzFile.parse_file(path)
            data = file.build()
            assert len(data) == file.file_header.len_file
            _ = anlz.AnlzFile.parse(data)



# coding: utf-8
#
# This code is part of rekordtools.
#
# Copyright (c) 2022, Dylan Jones

from .logger import logger
from .config import show_config, get_config
from .xml import RekordboxXml
from .anlz import get_anlz_paths, walk_anlz_paths, read_anlz_files, AnlzFile
from .mysettings import (
    get_mysetting_paths,
    read_mysetting_file,
    MySettingFile,
    MySetting2File,
    DjmMySettingFile,
    DevSettingFile,
)
from .db6 import Rekordbox6Database, open_rekordbox_database

try:
    from ._version import version as __version__
except ImportError:
    __version__ = "unknown"



# coding: utf-8
#
# This code is part of pyrekordbox.
#
# Copyright (c) 2022, Dylan Jones

import os
import pytest
from pyrekordbox.mysettings import (
    read_mysetting_file,
    MySettingFile,
    MySetting2File,
    DjmMySettingFile,
)
from pyrekordbox.mysettings.file import compute_checksum

TEST_ROOT = os.path.join(os.path.dirname(os.path.dirname(__file__)), ".testdata")

MYSETTING_KEYS = [
    "on_air_display",
    "lcd_brightness",
    "quantize",
    "auto_cue_level",
    "language",
    "jog_ring_brightness",
    "jog_ring_indicator",
    "slip_flashing",
    "eject_lock",
    "disc_slot_illumination",
    "sync",
    "play_mode",
    "quantize_beat_value",
    "hotcue_autoload",
    "hotcue_color",
    "needle_lock",
    "time_mode",
    "jog_mode",
    "auto_cue",
    "master_tempo",
    "tempo_range",
    "phase_meter",
]

MYSETTING2_KEYS = [
    "vinyl_speed_adjust",
    "jog_display_mode",
    "pad_button_brightness",
    "jog_lcd_brightness",
    "waveform_divisions",
    "waveform",
    "beat_jump_beat_value",
]

DJMMYSETTING_KEYS = [
    "channel_fader_curve",
    "cross_fader_curve",
    "headphones_pre_eq",
    "headphones_mono_split",
    "beat_fx_quantize",
    "mic_low_cut",
    "talk_over_mode",
    "talk_over_level",
    "midi_channel",
    "midi_button_type",
    "display_brightness",
    "indicator_brightness",
    "channel_fader_curve_long",
]


def _read_setting_file(type_, key, value):
    file = type_.upper() + ".DAT"
    root = os.path.join(TEST_ROOT, "mysettings", type_)
    path = os.path.join(root, key, value, file)
    if not os.path.exists(path):
        raise FileNotFoundError(path)
    return read_mysetting_file(path)


def _read_default_setting_file(type_):
    file = type_.upper() + ".DAT"
    root = os.path.join(TEST_ROOT, "mysettings")
    path = os.path.join(root, file)
    if not os.path.exists(path):
        raise FileNotFoundError(path)
    return read_mysetting_file(path)


def _get_values(type_, key):
    root = os.path.join(TEST_ROOT, "mysettings", type_, key)
    return os.listdir(root)


# -- MySetting -------------------------------------------------------------------------


@pytest.mark.parametrize("key", MYSETTING_KEYS)
def test_get_mysetting(key):
    for expected in _get_values("mysetting", key):
        file = _read_setting_file("mysetting", key, expected)
        assert file[key] == expected


def test_mysetting_defaults():
    file_default = _read_default_setting_file("mysetting")
    file = MySettingFile()
    for key in MYSETTING_KEYS:
        assert file[key] == file_default[key]


@pytest.mark.parametrize("key", MYSETTING_KEYS)
def test_set_mysetting(key):
    type_ = "mysetting"
    for val in _get_values(type_, key):
        # Read value file
        file_ref = _read_setting_file(type_, key, val)
        # Read default file and update value
        file = _read_default_setting_file(type_)
        file[key] = val
        # Check values are identical
        assert file[key] == file_ref[key]
        # Check binary data is identical
        assert file.build() == file_ref.build()


# -- MySetting2 ------------------------------------------------------------------------


@pytest.mark.parametrize("key", MYSETTING2_KEYS)
def test_get_mysetting2(key):
    for expected in _get_values("mysetting2", key):
        file = _read_setting_file("mysetting2", key, expected)
        assert file[key] == expected


def test_mysetting2_defaults():
    file_default = _read_default_setting_file("mysetting2")
    file = MySetting2File()
    for key in MYSETTING2_KEYS:
        assert file[key] == file_default[key]


@pytest.mark.parametrize("key", MYSETTING2_KEYS)
def test_set_mysetting2(key):
    type_ = "mysetting2"
    for val in _get_values(type_, key):
        # Read value file
        file_ref = _read_setting_file(type_, key, val)
        # Read default file and update value
        file = _read_default_setting_file(type_)
        file[key] = val
        # Check values are identical
        assert file[key] == file_ref[key]
        # Check binary data is identical
        assert file.build() == file_ref.build()


# -- DjmMySetting ----------------------------------------------------------------------


@pytest.mark.parametrize("key", DJMMYSETTING_KEYS)
def test_get_djmmysetting(key):
    for expected in _get_values("djmmysetting", key):
        file = _read_setting_file("djmmysetting", key, expected)
        assert file[key] == expected


def test_djmmysetting_defaults():
    file_default = _read_default_setting_file("djmmysetting")
    file = DjmMySettingFile()
    for key in DJMMYSETTING_KEYS:
        assert file[key] == file_default[key]


@pytest.mark.parametrize("key", DJMMYSETTING_KEYS)
def test_set_djmmysetting(key):
    type_ = "djmmysetting"
    for val in _get_values(type_, key):
        # Read value file
        file_ref = _read_setting_file(type_, key, val)
        # Read default file and update value
        file = _read_default_setting_file(type_)
        file[key] = val
        # Check values are identical
        assert file[key] == file_ref[key]
        # Check binary data is identical
        assert file.build() == file_ref.build()


# ======================================================================================


def test_mysetting_checksum():
    file_types = "MYSETTING.DAT", "MYSETTING2.DAT", "DJMMYSETTING.DAT"

    root = os.path.join(TEST_ROOT, "mysettings")
    for root, _, files in os.walk(root):
        for name in files:
            if name not in file_types:
                continue
            path = os.path.join(root, name)
            sett = read_mysetting_file(path)
            data = sett.build()
            checksum = compute_checksum(data, sett.struct)
            assert checksum == sett.parsed.checksum



# coding: utf-8
#
# This code is part of pyrekordbox.
#
# Copyright (c) 2022, Dylan Jones

"""This module contains common constants and methods used in other modules.

Contains all the path and settings handling of the Rekordbox installation(s) on
the users machine.
"""

import os
import sys
import json
import logging
import xml.etree.cElementTree as xml

logger = logging.getLogger(__name__)

# fmt: off
CRC16_XMODEM_TABLE = [
    0x0000, 0x1021, 0x2042, 0x3063, 0x4084, 0x50a5, 0x60c6, 0x70e7,
    0x8108, 0x9129, 0xa14a, 0xb16b, 0xc18c, 0xd1ad, 0xe1ce, 0xf1ef,
    0x1231, 0x0210, 0x3273, 0x2252, 0x52b5, 0x4294, 0x72f7, 0x62d6,
    0x9339, 0x8318, 0xb37b, 0xa35a, 0xd3bd, 0xc39c, 0xf3ff, 0xe3de,
    0x2462, 0x3443, 0x0420, 0x1401, 0x64e6, 0x74c7, 0x44a4, 0x5485,
    0xa56a, 0xb54b, 0x8528, 0x9509, 0xe5ee, 0xf5cf, 0xc5ac, 0xd58d,
    0x3653, 0x2672, 0x1611, 0x0630, 0x76d7, 0x66f6, 0x5695, 0x46b4,
    0xb75b, 0xa77a, 0x9719, 0x8738, 0xf7df, 0xe7fe, 0xd79d, 0xc7bc,
    0x48c4, 0x58e5, 0x6886, 0x78a7, 0x0840, 0x1861, 0x2802, 0x3823,
    0xc9cc, 0xd9ed, 0xe98e, 0xf9af, 0x8948, 0x9969, 0xa90a, 0xb92b,
    0x5af5, 0x4ad4, 0x7ab7, 0x6a96, 0x1a71, 0x0a50, 0x3a33, 0x2a12,
    0xdbfd, 0xcbdc, 0xfbbf, 0xeb9e, 0x9b79, 0x8b58, 0xbb3b, 0xab1a,
    0x6ca6, 0x7c87, 0x4ce4, 0x5cc5, 0x2c22, 0x3c03, 0x0c60, 0x1c41,
    0xedae, 0xfd8f, 0xcdec, 0xddcd, 0xad2a, 0xbd0b, 0x8d68, 0x9d49,
    0x7e97, 0x6eb6, 0x5ed5, 0x4ef4, 0x3e13, 0x2e32, 0x1e51, 0x0e70,
    0xff9f, 0xefbe, 0xdfdd, 0xcffc, 0xbf1b, 0xaf3a, 0x9f59, 0x8f78,
    0x9188, 0x81a9, 0xb1ca, 0xa1eb, 0xd10c, 0xc12d, 0xf14e, 0xe16f,
    0x1080, 0x00a1, 0x30c2, 0x20e3, 0x5004, 0x4025, 0x7046, 0x6067,
    0x83b9, 0x9398, 0xa3fb, 0xb3da, 0xc33d, 0xd31c, 0xe37f, 0xf35e,
    0x02b1, 0x1290, 0x22f3, 0x32d2, 0x4235, 0x5214, 0x6277, 0x7256,
    0xb5ea, 0xa5cb, 0x95a8, 0x8589, 0xf56e, 0xe54f, 0xd52c, 0xc50d,
    0x34e2, 0x24c3, 0x14a0, 0x0481, 0x7466, 0x6447, 0x5424, 0x4405,
    0xa7db, 0xb7fa, 0x8799, 0x97b8, 0xe75f, 0xf77e, 0xc71d, 0xd73c,
    0x26d3, 0x36f2, 0x0691, 0x16b0, 0x6657, 0x7676, 0x4615, 0x5634,
    0xd94c, 0xc96d, 0xf90e, 0xe92f, 0x99c8, 0x89e9, 0xb98a, 0xa9ab,
    0x5844, 0x4865, 0x7806, 0x6827, 0x18c0, 0x08e1, 0x3882, 0x28a3,
    0xcb7d, 0xdb5c, 0xeb3f, 0xfb1e, 0x8bf9, 0x9bd8, 0xabbb, 0xbb9a,
    0x4a75, 0x5a54, 0x6a37, 0x7a16, 0x0af1, 0x1ad0, 0x2ab3, 0x3a92,
    0xfd2e, 0xed0f, 0xdd6c, 0xcd4d, 0xbdaa, 0xad8b, 0x9de8, 0x8dc9,
    0x7c26, 0x6c07, 0x5c64, 0x4c45, 0x3ca2, 0x2c83, 0x1ce0, 0x0cc1,
    0xef1f, 0xff3e, 0xcf5d, 0xdf7c, 0xaf9b, 0xbfba, 0x8fd9, 0x9ff8,
    0x6e17, 0x7e36, 0x4e55, 0x5e74, 0x2e93, 0x3eb2, 0x0ed1, 0x1ef0,
]
# fmt: off


def _crc16(data, crc, table):
    """Calculate CRC16 using the given table.
    `data`      - data for calculating CRC, must be a string
    `crc`       - initial value
    `table`     - table for caclulating CRC (list of 256 integers)
    Return calculated value of CRC
    """
    for byte in data:
        crc = ((crc << 8) & 0xFF00) ^ table[((crc >> 8) & 0xFF) ^ byte]
    return crc & 0xFFFF


def crc16xmodem(data, crc=0):
    """Calculate CRC-CCITT (XModem) variant of CRC16.
    `data`      - data for calculating CRC, must be a string
    `crc`       - initial value
    Return calculated value of CRC
    """
    return _crc16(data, crc, CRC16_XMODEM_TABLE)


def _read_config_file(path):
    ext = os.path.splitext(path)[1].lower()
    if ext == ".cfg":
        from configparser import ConfigParser

        parser = ConfigParser()
        parser.read(path)
        return parser
    elif ext == ".toml":
        import toml

        return toml.load(path)
    elif ext in (".yaml", ".yml"):
        import yaml

        with open(path, "r") as stream:
            return yaml.safe_load(stream)
    return dict()


def read_pyrekordbox_configuration():
    """Reads the pyrekordbox configuration.

    So far only the `pioneer-install-dir` and `pioneer-app-dir` fileds in the
    `rekordbox` section are supported. Supported configuration files are
    pyproject.toml, setup.cfg, rekordbox.toml, rekordbox.cfg and rekordbox.yml.
    The files are searched for in that order.

    Returns
    -------
    pyrekordbox_config : dict
        The pyrekordbox configuration data, if found.
    """
    files = ["pyproject.toml", "setup.cfg"]
    for ext in [".toml", ".cfg", ".yaml", ".yml"]:
        files.append("rekordbox" + ext)

    for file in files:
        try:
            data = _read_config_file(file)
            config = dict(data["rekordbox"])
        except (ImportError, FileNotFoundError, KeyError):
            pass
        else:
            if config:
                return config
    return dict()


def get_pioneer_app_dir(path=""):
    """Returns the path of the Pioneer application data directory.

    On Windows, the Pioneer application data is stored in `/Users/user/AppData/Roaming`
    On macOS the application data is somewhere in `~/Libary/Application Support`.

    Parameters
    ----------
    path : str, optional
        If a path is given it will only be checked for validity. Otherwise,
        the default Pioneer directory will be constructed and checked.

    Returns
    -------
    pioneer_path : str
        The path to the Pioneer application data.
    """
    if not path:
        if sys.platform == "win32":
            # Windows: located in /Users/user/AppData/Roaming/
            app_data = os.environ["AppData"]
        elif sys.platform == "darwin":
            # MacOS: located in ~/Library/Application Support/
            home = os.path.expanduser("~")
            app_data = os.path.join(home, "Library", "Application Support")
        else:
            # Linux: not supported
            logger.warning(f"OS {sys.platform} not supported!")
            app_data = ""

        # Pioneer app data
        path = os.path.join(app_data, "Pioneer")

    path = os.path.abspath(path)
    if not os.path.exists(path):
        raise FileNotFoundError(f"The Pioneer app-dir {path} does not exist!")

    return path


def get_pioneer_install_dir(path=""):
    """Returns the path of the Pioneer program installation directory.

    On Windows, the Pioneer program data is stored in `/ProgramFiles/Pioneer`
    On macOS the program data is somewhere in `/Applications/`.

    Parameters
    ----------
    path : str, optional
        If a path is given it will only be checked for validity. Otherwise,
        the default Pioneer directory will be constructed and checked.

    Returns
    -------
    pioneer_path : str
        The path to the Pioneer program installation data.
    """
    if not path:
        if sys.platform == "win32":
            # Windows: located in /ProgramFiles/Pioneer
            program_files = os.environ["ProgramFiles"].replace("(x86)", "").strip()
            path = os.path.join(program_files, "Pioneer")
        elif sys.platform == "darwin":
            # MacOS: located in /Applications/
            path = os.path.abspath("/Applications")
        else:
            # Linux: not supported
            logger.warning(f"OS {sys.platform} not supported!")

    path = os.path.abspath(path)
    if not os.path.exists(path):
        raise FileNotFoundError(f"The Pioneer install-dir {path} does not exist!")

    return path


def _convert_type(s):
    # Try to parse as int, float, list of int, list of float
    types_ = int, float
    for type_ in types_:
        try:
            return type_(s)
        except ValueError:
            pass
        try:
            return [type_(x) for x in s.split(",")]
        except ValueError:
            pass

    return s


def read_rekordbox_settings(rekordbox_app_dir):
    """Finds and parses the 'rekordbox3.settings' file in the Rekordbox 5 or 6 app-dir.

    The settings file usually is called 'rekordbox3.settings' and is
    located in the application data directory of the corresponding Rekordbox
    (major) version.

    Parameters
    ----------
    rekordbox_app_dir : str
        The path of the application-data directory of Rekordbox 5 or 6.

    Returns
    -------
    settings : dict
        The parsed Rekordbox settings data.
    """
    # Get path of the settings file
    file_name = "rekordbox3.settings"
    path = os.path.join(rekordbox_app_dir, file_name)
    # if not os.path.exists(path):
    #    raise FileNotFoundError(f"No '{file_name}' file in '{rekordbox_app_dir}'.")

    # Parse the settings file
    settings = dict()
    tree = xml.parse(path)
    for element in tree.findall("VALUE"):
        name = element.attrib["name"]
        try:
            val = _convert_type(element.attrib["val"])
        except KeyError:
            device_setup = element.find("DEVICESETUP")
            val = {k: _convert_type(v) for k, v in device_setup.attrib.items()}
        settings[name] = val
    return settings


def read_rekordbox6_options(pioneer_app_dir):
    """Finds and parses the Rekordbox 6 `options.json` file with additional settings.

    The options file contains additional settings used by Rekordbox 6, for example the
    path of the new `master.db` database. It also contains some data nedded to open
    the database, which is encrypted using SQLCipher.

    Parameters
    ----------
    pioneer_app_dir : str, optional
        The path of the Pioneer application data.

    Returns
    -------
    options : dict
        The parsed rekordbox 6 options data.
    """
    # Get path of the options file
    opt_path = os.path.join(
        pioneer_app_dir, "rekordboxAgent", "storage", "options.json"
    )

    # Read and parse the options file
    with open(opt_path, "r") as fh:
        data = json.load(fh)
    options = dict()
    for key, value in data["options"]:
        options[key] = value

    return options


def read_rekordbox6_asar(rb6_install_dir):
    """Finds and parses the Rekordbox 6 `app.asar` archive file.

    An ASAR file is an archive used to package source code for an application
    using Electron. Rekordbox 6 stores some useful information for opening the
    new `master.db` database in this file.

    Parameters
    ----------
    rb6_install_dir : str
        The path of the Rekordbox 6 installation directory.

    Returns
    -------
    asar_data : str
        The data of the Rekordbox 6 `app.asar` archive file as ANSI encoded string.
    """
    # Find location of app asar file
    if sys.platform == "win32":
        location = os.path.join(
            rb6_install_dir, "rekordboxAgent-win32-x64", "resources"
        )
        encoding = "ANSI"
    elif sys.platform == "darwin":
        root = os.path.join(rb6_install_dir, "rekordbox.app", "Contents", "MacOS")
        location = os.path.join(root, "rekordboxAgent.app", "Contents", "Resources")
        encoding = "cp437"
    else:
        raise logger.warning(f"OS {sys.platform} not supported!")

    # Read asar file
    path = os.path.abspath(os.path.join(location, "app.asar"))
    with open(path, "rb") as fh:
        data = fh.read().decode(encoding)

    return data



# coding: utf-8
#
# This code is part of pyrekordbox.
#
# Copyright (c) 2022, Dylan Jones

import os
import re
from . import structs
from .file import FILES, MySettingFile, MySetting2File, DjmMySettingFile, DevSettingFile

RE_MYSETTING = re.compile(".*SETTING[0-9]?.DAT$")


def get_mysetting_paths(root, deep=False):
    files = list()
    for root, _, names in os.walk(root):
        for fname in names:
            if RE_MYSETTING.match(fname):
                files.append(os.path.join(root, fname))
        if not deep:
            break
    return files


def read_mysetting_file(path):
    obj = FILES[os.path.split(path)[1]]
    return obj.parse_file(path)



# coding: utf-8
#
# This code is part of pyrekordbox.
#
# Copyright (c) 2022, Dylan Jones

import re
from construct import Struct
from collections.abc import MutableMapping
from . import structs
from ..utils import crc16xmodem

RE_INVALID_KEY = re.compile("[_u][0-9]?", flags=re.IGNORECASE)


def compute_checksum(data, struct):
    """Computes the CRC16 XMODEM checksum for MySetting files.

    The checksum is calculated over the contents of the `data` field,
    except for `DJMSETTING.DAT` files where the checksum is calculated over all
    preceding bytes including the length fields.

    References
    ----------
    https://reveng.sourceforge.io/crc-catalogue/all.htm#crc.cat.crc-16-xmodem
    """
    start = 0 if struct == structs.DjmMySetting else 104
    return crc16xmodem(data[start:-4])


def _is_valid_key(k: str):
    return not RE_INVALID_KEY.match(k)


class SettingsFile(MutableMapping):
    """Abtract base class of Rekordbox settings files."""

    struct: Struct
    defaults: dict
    version: str = ""  # only used by DEVSETTING

    def __init__(self):
        super().__init__()
        self.parsed = None
        self.items = dict()

    @classmethod
    def parse(cls, data: bytes):
        """Parses the in-memory data of a Rekordbox settings binary file.

        Parameters
        ----------
        data : bytes
            The in-memory binary contents of a Rekordbox settings file.

        Returns
        -------
        self : SettingsFile
            The new instance with the parsed file content.
        """
        self = cls()
        self._parse(data)
        return self

    @classmethod
    def parse_file(cls, path: str):
        """Reads and parses a Rekordbox settings binary file.

        Parameters
        ----------
        path : str
            The path of a Rekordbox settings file which is used to read
            the file contents before parsing the binary data.

        Returns
        -------
        self : AnlzFile
            The new instance with the parsed file content.

        See Also
        --------
        SettingsFile.parse: Parses the data of a Rekordbox settings file.
        """
        with open(path, "rb") as fh:
            data = fh.read()
        return cls.parse(data)

    def _parse(self, data):
        parsed = self.struct.parse(data)
        keys = filter(_is_valid_key, parsed.data.keys())
        items = dict()
        for key in keys:
            items[key] = str(parsed.data[key])

        self.parsed = parsed
        self.items.update(items)

    def build(self):
        # Copy defaults and update with cuirrent data
        items = self.defaults.copy()
        items.update(self.items)

        # Create file data
        file_items = {"data": items, "checksum": 0}
        if self.version:
            file_items["version"] = self.version

        # Compute and update checksum. For this the data has to be serialized twice:
        # Once for generating the checksum and another time for writing the data
        # with the updated checksum
        data = self.struct.build(file_items)
        checksum = compute_checksum(data, self.struct)
        file_items["checksum"] = checksum

        # Write data with updated checksum
        return self.struct.build(file_items)

    def save(self, path):
        data = self.build()
        with open(path, "wb") as fh:
            fh.write(data)

    def __len__(self):
        return len(self.defaults.keys())

    def __iter__(self):
        return iter(self.defaults)

    def __getitem__(self, key):
        try:
            return self.items[key]
        except KeyError:
            return self.defaults[key]

    def __setitem__(self, key, value):
        if key not in self.defaults.keys():
            raise KeyError(f"Key {key} not a valid field of {self.__class__.__name__}")
        self.items[key] = value

    def __delitem__(self, key):
        del self.items[key]

    def get(self, key, default=None):
        try:
            return self.__getitem__(key)
        except KeyError:
            return default

    def set(self, key, value):
        self.__setitem__(key, value)

    def __repr__(self):
        return f"{self.__class__.__name__}()"


class DevSettingFile(SettingsFile):

    struct = structs.DevSetting
    defaults = dict(entries="")


class DjmMySettingFile(SettingsFile):

    struct = structs.DjmMySetting
    defaults = {
        "channel_fader_curve": structs.ChannelFaderCurve.linear,
        "cross_fader_curve": structs.CrossfaderCurve.fast_cut,
        "headphones_pre_eq": structs.HeadphonesPreEQ.post_eq,
        "headphones_mono_split": structs.HeadphonesMonoSplit.stereo,
        "beat_fx_quantize": structs.BeatFXQuantize.on,
        "mic_low_cut": structs.MicLowCut.on,
        "talk_over_mode": structs.TalkOverMode.advanced,
        "talk_over_level": structs.TalkOverLevel.minus_18db,
        "midi_channel": structs.MidiChannel.one,
        "midi_button_type": structs.MidiButtonType.toggle,
        "display_brightness": structs.MixerDisplayBrightness.five,
        "indicator_brightness": structs.MixerIndicatorBrightness.three,
        "channel_fader_curve_long": structs.ChannelFaderCurveLong.exponential,
    }


class MySettingFile(SettingsFile):

    struct = structs.MySetting
    defaults = {
        "auto_cue": structs.AutoCue.on,
        "auto_cue_level": structs.AutoCueLevel.memory,
        "disc_slot_illumination": structs.DiscSlotIllumination.bright,
        "eject_lock": structs.EjectLock.unlock,
        "hotcue_autoload": structs.HotCueAutoLoad.on,
        "hotcue_color": structs.HotCueColor.off,
        "jog_mode": structs.JogMode.vinyl,
        "jog_ring_brightness": structs.JogRingBrightness.bright,
        "jog_ring_indicator": structs.JogRingIndicator.on,
        "language": structs.Language.english,
        "lcd_brightness": structs.LCDBrightness.three,
        "master_tempo": structs.MasterTempo.off,
        "needle_lock": structs.NeedleLock.lock,
        "on_air_display": structs.OnAirDisplay.on,
        "phase_meter": structs.PhaseMeter.type1,
        "play_mode": structs.PlayMode.single,
        "quantize": structs.Quantize.on,
        "quantize_beat_value": structs.QuantizeBeatValue.one,
        "slip_flashing": structs.SlipFlashing.on,
        "sync": structs.Sync.off,
        "tempo_range": structs.TempoRange.ten,
        "time_mode": structs.TimeMode.remain,
    }


class MySetting2File(SettingsFile):

    struct = structs.MySetting2
    defaults = {
        "vinyl_speed_adjust": structs.VinylSpeedAdjust.touch,
        "jog_display_mode": structs.JogDisplayMode.auto,
        "pad_button_brightness": structs.PadButtonBrightness.three,
        "jog_lcd_brightness": structs.JogLCDBrightness.three,
        "waveform_divisions": structs.WaveformDivisions.phrase,
        "waveform": structs.Waveform.waveform,
        "beat_jump_beat_value": structs.BeatJumpBeatValue.sixteen,
    }


FILES = {
    "DEVSETTING.DAT": DevSettingFile,
    "DJMMYSETTING.DAT": DjmMySettingFile,
    "MYSETTING.DAT": MySettingFile,
    "MYSETTING2.DAT": MySetting2File,
}



# coding: utf-8
#
# This code is part of pyrekordbox.
#
# Copyright (c) 2022, Dylan Jones

from .tables import (
    AgentRegistry,
    CloudAgentRegistry,
    ContentActiveCensor,
    ContentCue,
    ContentFile,
    DjmdActiveCensor,
    DjmdAlbum,
    DjmdArtist,
    DjmdCategory,
    DjmdColor,
    DjmdContent,
    DjmdCue,
    DjmdDevice,
    DjmdGenre,
    DjmdHistory,
    DjmdHotCueBanklist,
    DjmdKey,
    DjmdLabel,
    DjmdMenuItems,
    DjmdMixerParam,
    DjmdMyTag,
    DjmdPlaylist,
    DjmdProperty,
    DjmdRelatedTracks,
    DjmdSampler,
    DjmdSongHistory,
    DjmdSongHotCueBanklist,
    DjmdSongMyTag,
    DjmdSongPlaylist,
    DjmdSongRelatedTracks,
    DjmdSongSampler,
    DjmdSongTagList,
    DjmdSort,
    HotCueBanklistCue,
    ImageFile,
    SettingFile,
    UuidIDMap,
)

from .database import Rekordbox6Database, open_rekordbox_database



# coding: utf-8
#
# This code is part of pyrekordbox.
#
# Copyright (c) 2022, Dylan Jones

"""Rekordbox 6 `master.db` SQLAlchemy tabel declarations."""

from sqlalchemy import Column, Integer, VARCHAR, BigInteger, SmallInteger, DateTime
from sqlalchemy import Text, ForeignKey, Float
from sqlalchemy.orm import declarative_base, relationship

__all__ = [
    "AgentRegistry",
    "CloudAgentRegistry",
    "ContentActiveCensor",
    "ContentCue",
    "ContentFile",
    "DjmdActiveCensor",
    "DjmdAlbum",
    "DjmdArtist",
    "DjmdCategory",
    "DjmdColor",
    "DjmdContent",
    "DjmdCue",
    "DjmdDevice",
    "DjmdGenre",
    "DjmdHistory",
    "DjmdHotCueBanklist",
    "DjmdKey",
    "DjmdLabel",
    "DjmdMenuItems",
    "DjmdMixerParam",
    "DjmdMyTag",
    "DjmdPlaylist",
    "DjmdProperty",
    "DjmdRelatedTracks",
    "DjmdSampler",
    "DjmdSongHistory",
    "DjmdSongHotCueBanklist",
    "DjmdSongMyTag",
    "DjmdSongPlaylist",
    "DjmdSongRelatedTracks",
    "DjmdSongSampler",
    "DjmdSongTagList",
    "DjmdSort",
    "HotCueBanklistCue",
    "ImageFile",
    "SettingFile",
    "UuidIDMap",
]


# -- Base- and Mixin classes -----------------------------------------------------------


Base = declarative_base()


class StatsTime:

    created_at = Column(DateTime, nullable=False)
    updated_at = Column(DateTime, nullable=False)


class StatsFull:

    ID: Column

    UUID = Column(VARCHAR(255), default=None)
    rb_data_status = Column(Integer, default=0)
    rb_local_data_status = Column(Integer, default=0)
    rb_local_deleted = Column(SmallInteger, default=0)
    rb_local_synced = Column(SmallInteger, default=0)
    usn = Column(BigInteger, default=None)
    rb_local_usn = Column(BigInteger, default=None)
    created_at = Column(DateTime, nullable=False)
    updated_at = Column(DateTime, nullable=False)

    def __repr__(self):
        return f"<{self.__class__.__name__}({self.ID})>"


# -- Table declarations ----------------------------------------------------------------


class AgentRegistry(Base, StatsTime):

    __tablename__ = "agentRegistry"

    registry_id = Column(VARCHAR(255), primary_key=True)
    id_1 = Column(VARCHAR(255), default=None)
    id_2 = Column(VARCHAR(255), default=None)
    int_1 = Column(Integer, default=None)
    int_2 = Column(Integer, default=None)
    str_1 = Column(VARCHAR(255), default=None)
    str_2 = Column(VARCHAR(255), default=None)
    date_1 = Column(DateTime, default=None)
    date_2 = Column(DateTime, default=None)
    text_1 = Column(Text, default=None)
    text_2 = Column(Text, default=None)


class CloudAgentRegistry(Base, StatsFull):

    __tablename__ = "cloudAgentRegistry"

    ID = Column(VARCHAR(255), primary_key=True)
    int_1 = Column(Integer, default=None)
    int_2 = Column(Integer, default=None)
    str_1 = Column(VARCHAR(255), default=None)
    str_2 = Column(VARCHAR(255), default=None)
    date_1 = Column(DateTime, default=None)
    date_2 = Column(DateTime, default=None)
    text_1 = Column(Text, default=None)
    text_2 = Column(Text, default=None)


class ContentActiveCensor(Base, StatsFull):

    __tablename__ = "contentActiveCensor"

    ID = Column(VARCHAR(255), primary_key=True)
    ContentID = Column(VARCHAR(255), ForeignKey("djmdContent.ID"), default=None)
    ActiveCensors = Column(Text, default=None)
    rb_activecensor_count = Column(Integer, default=None)


class ContentCue(Base, StatsFull):

    __tablename__ = "contentCue"

    ID = Column(VARCHAR(255), primary_key=True)
    ContentID = Column(VARCHAR(255), ForeignKey("djmdContent.ID"), default=None)
    Cues = Column(Text, default=None)
    rb_cue_count = Column(Integer, default=None)


class ContentFile(Base, StatsFull):

    __tablename__ = "contentFile"

    ID = Column(VARCHAR(255), primary_key=True)
    ContentID = Column(VARCHAR(255), ForeignKey("djmdContent.ID"), default=None)
    Path = Column(VARCHAR(255), default=None)
    Hash = Column(VARCHAR(255), default=None)
    Size = Column(Integer, default=None)
    rb_local_path = Column(VARCHAR(255), default=None)
    rb_insync_hash = Column(VARCHAR(255), default=None)
    rb_insync_local_usn = Column(BigInteger, default=None)
    rb_file_hash_dirty = Column(Integer, default=0)
    rb_local_file_status = Column(Integer, default=0)
    rb_in_progress = Column(SmallInteger, default=0)
    rb_process_type = Column(Integer, default=0)
    rb_temp_path = Column(VARCHAR(255), default=None)
    rb_priority = Column(Integer, default=50)
    rb_file_size_dirty = Column(Integer, default=0)


class DjmdActiveCensor(Base, StatsFull):

    __tablename__ = "djmdActiveCensor"

    ID = Column(VARCHAR(255), primary_key=True)
    ContentID = Column(VARCHAR(255), default=None)
    InMsec = Column(Integer, default=None)
    OutMsec = Column(Integer, default=None)
    Info = Column(Integer, default=None)
    ParameterList = Column(Text, default=None)
    ContentUUID = Column(VARCHAR(255), default=None)


class DjmdAlbum(Base, StatsFull):

    __tablename__ = "djmdAlbum"

    ID = Column(VARCHAR(255), primary_key=True)
    Name = Column(VARCHAR(255), default=None)
    AlbumArtistID = Column(VARCHAR(255), default=None)
    ImagePath = Column(VARCHAR(255), default=None)
    Compilation = Column(Integer, default=None)
    SearchStr = Column(VARCHAR(255), default=None)

    def __repr__(self):
        s = f"{self.ID: <10} Name={self.Name}"
        return f"<{self.__class__.__name__}({s})>"


class DjmdArtist(Base, StatsFull):

    __tablename__ = "djmdArtist"

    ID = Column(VARCHAR(255), primary_key=True)
    Name = Column(VARCHAR(255), default=None)
    SearchStr = Column(VARCHAR(255), default=None)

    def __repr__(self):
        s = f"{self.ID: <10} Name={self.Name}"
        return f"<{self.__class__.__name__}({s})>"


class DjmdCategory(Base, StatsFull):

    __tablename__ = "djmdCategory"

    ID = Column(VARCHAR(255), primary_key=True)
    MenuItemID = Column(VARCHAR(255), default=None)
    Seq = Column(Integer, default=None)
    Disable = Column(Integer, default=None)
    InfoOrder = Column(Integer, default=None)


class DjmdColor(Base, StatsFull):

    __tablename__ = "djmdColor"

    ID = Column(VARCHAR(255), primary_key=True)
    ColorCode = Column(Integer, default=None)
    SortKey = Column(Integer, default=None)
    Commnt = Column(VARCHAR(255), default=None)

    def __repr__(self):
        s = f"{self.ID: <2} Comment={self.Commnt}"
        return f"<{self.__class__.__name__}({s})>"


class DjmdContent(Base, StatsFull):

    __tablename__ = "djmdContent"

    ID = Column(VARCHAR(255), primary_key=True)
    FolderPath = Column(VARCHAR(255), default=None)
    FileNameL = Column(VARCHAR(255), default=None)
    FileNameS = Column(VARCHAR(255), default=None)
    Title = Column(VARCHAR(255), default=None)
    ArtistID = Column(VARCHAR(255), ForeignKey("djmdArtist.ID"), default=None)
    AlbumID = Column(VARCHAR(255), ForeignKey("djmdAlbum.ID"), default=None)
    GenreID = Column(VARCHAR(255), ForeignKey("djmdGenre.ID"), default=None)
    BPM = Column(Integer, default=None)
    Length = Column(Integer, default=None)
    TrackNo = Column(Integer, default=None)
    BitRate = Column(Integer, default=None)
    BitDepth = Column(Integer, default=None)
    Commnt = Column(Text, default=None)
    FileType = Column(Integer, default=None)
    Rating = Column(Integer, default=None)
    ReleaseYear = Column(Integer, default=None)
    RemixerID = Column(VARCHAR(255), ForeignKey("djmdArtist.ID"), default=None)
    LabelID = Column(VARCHAR(255), ForeignKey("djmdLabel.ID"), default=None)
    OrgArtistID = Column(VARCHAR(255), ForeignKey("djmdArtist.ID"), default=None)
    KeyID = Column(VARCHAR(255), ForeignKey("djmdKey.ID"), default=None)
    StockDate = Column(VARCHAR(255), default=None)
    ColorID = Column(VARCHAR(255), ForeignKey("djmdColor.ID"), default=None)
    DJPlayCount = Column(VARCHAR(255), default=None)
    ImagePath = Column(VARCHAR(255), default=None)
    MasterDBID = Column(VARCHAR(255), default=None)
    MasterSongID = Column(VARCHAR(255), default=None)
    AnalysisDataPath = Column(VARCHAR(255), default=None)
    SearchStr = Column(VARCHAR(255), default=None)
    FileSize = Column(Integer, default=None)
    DiscNo = Column(Integer, default=None)
    ComposerID = Column(VARCHAR(255), ForeignKey("djmdArtist.ID"), default=None)
    Subtitle = Column(VARCHAR(255), default=None)
    SampleRate = Column(Integer, default=None)
    DisableQuantize = Column(Integer, default=None)
    Analysed = Column(Integer, default=None)
    ReleaseDate = Column(VARCHAR(255), default=None)
    DateCreated = Column(VARCHAR(255), default=None)
    ContentLink = Column(Integer, default=None)
    Tag = Column(VARCHAR(255), default=None)
    ModifiedByRBM = Column(VARCHAR(255), default=None)
    HotCueAutoLoad = Column(VARCHAR(255), default=None)
    DeliveryControl = Column(VARCHAR(255), default=None)
    DeliveryComment = Column(VARCHAR(255), default=None)
    CueUpdated = Column(VARCHAR(255), default=None)
    AnalysisUpdated = Column(VARCHAR(255), default=None)
    TrackInfoUpdated = Column(VARCHAR(255), default=None)
    Lyricist = Column(VARCHAR(255), ForeignKey("djmdArtist.ID"), default=None)
    ISRC = Column(VARCHAR(255), default=None)
    SamplerTrackInfo = Column(Integer, default=None)
    SamplerPlayOffset = Column(Integer, default=None)
    SamplerGain = Column(Float, default=None)
    VideoAssociate = Column(VARCHAR(255), default=None)
    LyricStatus = Column(Integer, default=None)
    ServiceID = Column(Integer, default=None)
    OrgFolderPath = Column(VARCHAR(255), default=None)
    Reserved1 = Column(Text, default=None)
    Reserved2 = Column(Text, default=None)
    Reserved3 = Column(Text, default=None)
    Reserved4 = Column(Text, default=None)
    ExtInfo = Column(Text, default=None)
    rb_file_id = Column(VARCHAR(255), default=None)
    DeviceID = Column(VARCHAR(255), default=None)
    rb_LocalFolderPath = Column(VARCHAR(255), default=None)
    SrcID = Column(VARCHAR(255), default=None)
    SrcTitle = Column(VARCHAR(255), default=None)
    SrcArtistName = Column(VARCHAR(255), default=None)
    SrcAlbumName = Column(VARCHAR(255), default=None)
    SrcLength = Column(Integer, default=None)

    Artist = relationship("DjmdArtist", foreign_keys=[ArtistID])
    Album = relationship("DjmdAlbum", foreign_keys=[AlbumID])
    Genre = relationship("DjmdGenre", foreign_keys=[GenreID])
    Remixer = relationship("DjmdArtist", foreign_keys=[RemixerID])
    Label = relationship("DjmdLabel", foreign_keys=[LabelID])
    OrgArtist = relationship("DjmdArtist", foreign_keys=[OrgArtistID])
    Key = relationship("DjmdKey", foreign_keys=[KeyID])
    Color = relationship("DjmdColor", foreign_keys=[ColorID])
    Composer = relationship("DjmdArtist", foreign_keys=[ComposerID])

    def __repr__(self):
        s = f"{self.ID: <10} Title={self.Title}"
        return f"<{self.__class__.__name__}({s})>"


class DjmdCue(Base, StatsFull):

    __tablename__ = "djmdCue"

    ID = Column(VARCHAR(255), primary_key=True)
    ContentID = Column(VARCHAR(255), ForeignKey("djmdContent.ID"), default=None)
    InMsec = Column(Integer, default=None)
    InFrame = Column(Integer, default=None)
    InMpegFrame = Column(Integer, default=None)
    InMpegAbs = Column(Integer, default=None)
    OutMsec = Column(Integer, default=None)
    OutFrame = Column(Integer, default=None)
    OutMpegFrame = Column(Integer, default=None)
    OutMpegAbs = Column(Integer, default=None)
    Kind = Column(Integer, default=None)
    Color = Column(Integer, default=None)
    ColorTableIndex = Column(Integer, default=None)
    ActiveLoop = Column(Integer, default=None)
    Comment = Column(VARCHAR(255), default=None)
    BeatLoopSize = Column(Integer, default=None)
    CueMicrosec = Column(Integer, default=None)
    InPointSeekInfo = Column(VARCHAR(255), default=None)
    OutPointSeekInfo = Column(VARCHAR(255), default=None)
    ContentUUID = Column(VARCHAR(255), ForeignKey("djmdContent.UUID"), default=None)


class DjmdDevice(Base, StatsFull):

    __tablename__ = "djmdDevice"

    ID = Column(VARCHAR(255), primary_key=True)
    MasterDBID = Column(VARCHAR(255), default=None)
    Name = Column(VARCHAR(255), default=None)

    def __repr__(self):
        s = f"{self.ID: <2} Name={self.Name}"
        return f"<{self.__class__.__name__}({s})>"


class DjmdGenre(Base, StatsFull):

    __tablename__ = "djmdGenre"

    ID = Column(VARCHAR(255), primary_key=True)
    Name = Column(VARCHAR(255), default=None)

    def __repr__(self):
        s = f"{self.ID: <2} Name={self.Name}"
        return f"<{self.__class__.__name__}({s})>"


class DjmdHistory(Base, StatsFull):

    __tablename__ = "djmdHistory"

    ID = Column(VARCHAR(255), primary_key=True)
    Seq = Column(Integer, default=None)
    Name = Column(VARCHAR(255), default=None)
    Attribute = Column(Integer, default=None)
    ParentID = Column(VARCHAR(255), ForeignKey("djmdHistory.ID"), default=None)
    DateCreated = Column(VARCHAR(255), default=None)

    Songs = relationship("DjmdSongHistory", back_populates="History")

    def __repr__(self):
        s = f"{self.ID: <2} Name={self.Name}"
        return f"<{self.__class__.__name__}({s})>"


class DjmdSongHistory(Base, StatsFull):

    __tablename__ = "djmdSongHistory"

    ID = Column(VARCHAR(255), primary_key=True)
    HistoryID = Column(VARCHAR(255), ForeignKey("djmdHistory.ID"), default=None)
    ContentID = Column(VARCHAR(255), ForeignKey("djmdContent.ID"), default=None)
    TrackNo = Column(Integer, default=None)

    History = relationship("DjmdHistory", back_populates="Songs")


class DjmdHotCueBanklist(Base, StatsFull):

    __tablename__ = "djmdHotCueBanklist"

    ID = Column(VARCHAR(255), primary_key=True)
    Seq = Column(Integer, default=None)
    Name = Column(VARCHAR(255), default=None)
    ImagePath = Column(VARCHAR(255), default=None)
    Attribute = Column(Integer, default=None)
    ParentID = Column(VARCHAR(255), ForeignKey("djmdHotCueBanklist.ID"), default=None)

    def __repr__(self):
        s = f"{self.ID: <2} Name={self.Name}"
        return f"<{self.__class__.__name__}({s})>"


class DjmdSongHotCueBanklist(Base, StatsFull):

    __tablename__ = "djmdSongHotCueBanklist"

    ID = Column(VARCHAR(255), primary_key=True)
    HotCueBanklistID = Column(
        VARCHAR(255), ForeignKey("djmdHotCueBanklist.ID"), default=None
    )
    ContentID = Column(VARCHAR(255), ForeignKey("djmdContent.ID"), default=None)
    TrackNo = Column(Integer, default=None)
    CueID = Column(VARCHAR(255), default=None)
    InMsec = Column(Integer, default=None)
    InFrame = Column(Integer, default=None)
    InMpegFrame = Column(Integer, default=None)
    InMpegAbs = Column(Integer, default=None)
    OutMsec = Column(Integer, default=None)
    OutFrame = Column(Integer, default=None)
    OutMpegFrame = Column(Integer, default=None)
    OutMpegAbs = Column(Integer, default=None)
    Color = Column(Integer, default=None)
    ColorTableIndex = Column(Integer, default=None)
    ActiveLoop = Column(Integer, default=None)
    Comment = Column(VARCHAR(255), default=None)
    BeatLoopSize = Column(Integer, default=None)
    CueMicrosec = Column(Integer, default=None)
    InPointSeekInfo = Column(VARCHAR(255), default=None)
    OutPointSeekInfo = Column(VARCHAR(255), default=None)
    HotCueBanklistUUID = Column(
        VARCHAR(255), ForeignKey("djmdHotCueBanklist.UUID"), default=None
    )


class DjmdKey(Base, StatsFull):

    __tablename__ = "djmdKey"

    ID = Column(VARCHAR(255), primary_key=True)
    ScaleName = Column(VARCHAR(255), default=None)
    Seq = Column(Integer, default=None)

    def __repr__(self):
        s = f"{self.ID: <2} Name={self.ScaleName}"
        return f"<{self.__class__.__name__}({s})>"


class DjmdLabel(Base, StatsFull):

    __tablename__ = "djmdLabel"

    ID = Column(VARCHAR(255), primary_key=True)
    Name = Column(VARCHAR(255), default=None)

    def __repr__(self):
        s = f"{self.ID: <2} Name={self.Name}"
        return f"<{self.__class__.__name__}({s})>"


class DjmdMenuItems(Base, StatsFull):

    __tablename__ = "djmdMenuItems"

    ID = Column(VARCHAR(255), primary_key=True)
    Class = Column(Integer, default=None)
    Name = Column(VARCHAR(255), default=None)

    def __repr__(self):
        s = f"{self.ID: <2} Name={self.Name}"
        return f"<{self.__class__.__name__}({s})>"


class DjmdMixerParam(Base, StatsFull):

    __tablename__ = "djmdMixerParam"

    ID = Column(VARCHAR(255), primary_key=True)
    ContentID = Column(VARCHAR(255), ForeignKey("djmdContent.ID"), default=None)
    GainHigh = Column(Integer, default=None)
    GainLow = Column(Integer, default=None)
    PeakHigh = Column(Integer, default=None)
    PeakLow = Column(Integer, default=None)


class DjmdMyTag(Base, StatsFull):

    __tablename__ = "djmdMyTag"

    ID = Column(VARCHAR(255), primary_key=True)
    Seq = Column(Integer, default=None)
    Name = Column(VARCHAR(255), default=None)
    Attribute = Column(Integer, default=None)
    ParentID = Column(VARCHAR(255), ForeignKey("djmdMyTag.ID"), default=None)

    MyTags = relationship("DjmdSongMyTag", back_populates="MyTag")

    def __repr__(self):
        s = f"{self.ID: <2} Name={self.Name}"
        return f"<{self.__class__.__name__}({s})>"


class DjmdSongMyTag(Base, StatsFull):

    __tablename__ = "djmdSongMyTag"

    ID = Column(VARCHAR(255), primary_key=True)
    MyTagID = Column(VARCHAR(255), ForeignKey("djmdMyTag.ID"), default=None)
    ContentID = Column(VARCHAR(255), ForeignKey("djmdContent.ID"), default=None)
    TrackNo = Column(Integer, default=None)

    MyTag = relationship("DjmdMyTag", back_populates="MyTags")


class DjmdPlaylist(Base, StatsFull):

    __tablename__ = "djmdPlaylist"

    ID = Column(VARCHAR(255), primary_key=True)
    Seq = Column(Integer, default=None)
    Name = Column(VARCHAR(255), default=None)
    ImagePath = Column(VARCHAR(255), default=None)
    Attribute = Column(Integer, default=None)
    ParentID = Column(VARCHAR(255), ForeignKey("djmdMyTag.ID"), default=None)
    SmartList = Column(Text, default=None)

    Songs = relationship("DjmdSongPlaylist", back_populates="Playlist")

    def __repr__(self):
        s = f"{self.ID: <2} Name={self.Name}"
        return f"<{self.__class__.__name__}({s})>"


class DjmdSongPlaylist(Base, StatsFull):

    __tablename__ = "djmdSongPlaylist"

    ID = Column(VARCHAR(255), primary_key=True)
    PlaylistID = Column(VARCHAR(255), ForeignKey("djmdPlaylist.ID"), default=None)
    ContentID = Column(VARCHAR(255), ForeignKey("djmdContent.ID"), default=None)
    TrackNo = Column(Integer, default=None)

    Playlist = relationship("DjmdPlaylist", back_populates="Songs")


class DjmdRelatedTracks(Base, StatsFull):

    __tablename__ = "djmdRelatedTracks"

    ID = Column(VARCHAR(255), primary_key=True)
    Seq = Column(Integer, default=None)
    Name = Column(VARCHAR(255), default=None)
    Attribute = Column(Integer, default=None)
    ParentID = Column(VARCHAR(255), ForeignKey("djmdRelatedTracks.ID"), default=None)
    Criteria = Column(Text, default=None)

    Songs = relationship("DjmdSongRelatedTracks", back_populates="RelatedTracks")

    def __repr__(self):
        s = f"{self.ID: <2} Name={self.Name}"
        return f"<{self.__class__.__name__}({s})>"


class DjmdSongRelatedTracks(Base, StatsFull):

    __tablename__ = "djmdSongRelatedTracks"

    ID = Column(VARCHAR(255), primary_key=True)
    RelatedTracksID = Column(
        VARCHAR(255), ForeignKey("djmdRelatedTracks.ID"), default=None
    )
    ContentID = Column(VARCHAR(255), ForeignKey("djmdContent.ID"), default=None)
    TrackNo = Column(Integer, default=None)

    RelatedTracks = relationship("DjmdRelatedTracks", back_populates="Songs")


class DjmdSampler(Base, StatsFull):

    __tablename__ = "djmdSampler"

    ID = Column(VARCHAR(255), primary_key=True)
    Seq = Column(Integer, default=None)
    Name = Column(VARCHAR(255), default=None)
    Attribute = Column(Integer, default=None)
    ParentID = Column(VARCHAR(255), ForeignKey("djmdSampler.ID"), default=None)

    Songs = relationship("DjmdSongSampler", back_populates="Sampler")

    def __repr__(self):
        s = f"{self.ID: <2} Name={self.Name}"
        return f"<{self.__class__.__name__}({s})>"


class DjmdSongSampler(Base, StatsFull):

    __tablename__ = "djmdSongSampler"

    ID = Column(VARCHAR(255), primary_key=True)
    SamplerID = Column(VARCHAR(255), ForeignKey("djmdSampler.ID"), default=None)
    ContentID = Column(VARCHAR(255), ForeignKey("djmdContent.ID"), default=None)
    TrackNo = Column(Integer, default=None)

    Sampler = relationship("DjmdSampler", back_populates="Songs")


class DjmdSongTagList(Base, StatsFull):

    __tablename__ = "djmdSongTagList"

    ID = Column(VARCHAR(255), primary_key=True)
    ContentID = Column(VARCHAR(255), ForeignKey("djmdContent.ID"), default=None)
    TrackNo = Column(Integer, default=None)


class DjmdSort(Base, StatsFull):

    __tablename__ = "djmdSort"

    ID = Column(VARCHAR(255), primary_key=True)
    MenuItemID = Column(VARCHAR(255), default=None)
    Seq = Column(Integer, default=None)
    Disable = Column(Integer, default=None)


class HotCueBanklistCue(Base, StatsFull):

    __tablename__ = "hotCueBanklistCue"

    ID = Column(VARCHAR(255), primary_key=True)
    HotCueBanklistID = Column(VARCHAR(255), default=None)
    Cues = Column(Text, default=None)
    rb_cue_count = Column(Integer, default=None)


class DjmdProperty(Base, StatsTime):

    __tablename__ = "djmdProperty"

    DBID = Column(VARCHAR(255), primary_key=True)
    DBVersion = Column(VARCHAR(255), default=None)
    BaseDBDrive = Column(VARCHAR(255), default=None)
    CurrentDBDrive = Column(VARCHAR(255), default=None)
    DeviceID = Column(VARCHAR(255), default=None)
    Reserved1 = Column(Text, default=None)
    Reserved2 = Column(Text, default=None)
    Reserved3 = Column(Text, default=None)
    Reserved4 = Column(Text, default=None)
    Reserved5 = Column(Text, default=None)


class ImageFile(Base, StatsFull):

    __tablename__ = "imageFile"

    ID = Column(VARCHAR(255), primary_key=True)
    TableName = Column(VARCHAR(255), default=None)
    TargetUUID = Column(VARCHAR(255), default=None)
    TargetID = Column(VARCHAR(255), default=None)
    Path = Column(VARCHAR(255), default=None)
    Hash = Column(VARCHAR(255), default=None)
    Size = Column(Integer, default=None)
    rb_local_path = Column(VARCHAR(255), default=None)
    rb_insync_hash = Column(VARCHAR(255), default=None)
    rb_insync_local_usn = Column(BigInteger, default=None)
    rb_file_hash_dirty = Column(Integer, default=0)
    rb_local_file_status = Column(Integer, default=0)
    rb_in_progress = Column(SmallInteger, default=0)
    rb_process_type = Column(Integer, default=0)
    rb_temp_path = Column(VARCHAR(255), default=None)
    rb_priority = Column(Integer, default=50)
    rb_file_size_dirty = Column(Integer, default=0)


class SettingFile(Base, StatsFull):

    __tablename__ = "settingFile"

    ID = Column(VARCHAR(255), primary_key=True)
    Path = Column(VARCHAR(255), default=None)
    Hash = Column(VARCHAR(255), default=None)
    Size = Column(Integer, default=None)
    rb_local_path = Column(VARCHAR(255), default=None)
    rb_insync_hash = Column(VARCHAR(255), default=None)
    rb_insync_local_usn = Column(BigInteger, default=None)
    rb_file_hash_dirty = Column(Integer, default=0)
    rb_file_size_dirty = Column(Integer, default=0)


class UuidIDMap(Base, StatsFull):

    __tablename__ = "uuidIDMap"

    ID = Column(VARCHAR(255), primary_key=True)
    TableName = Column(VARCHAR(255), default=None)
    TargetUUID = Column(VARCHAR(255), default=None)
    CurrentID = Column(VARCHAR(255), default=None)



# coding: utf-8
#
# This code is part of pyrekordbox.
#
# Copyright (c) 2022, Dylan Jones

"""Package logger configuration."""

import logging

logger = logging.getLogger("pyrekordbox")

# Logging format
frmt = "[%(asctime)s] %(name)s:%(levelname)-8s - %(message)s"
formatter = logging.Formatter(frmt, datefmt="%H:%M:%S")

# Set up console logger
sh = logging.StreamHandler()
sh.setLevel(logging.DEBUG)
sh.setFormatter(formatter)
logger.addHandler(sh)

# Set logging level
logger.setLevel(logging.WARNING)
logging.root.setLevel(logging.NOTSET)



# coding: utf-8
#
# This code is part of pyrekordbox.
#
# Copyright (c) 2022, Dylan Jones

import os
import re
import base64
import warnings
import blowfish
import logging
import sys
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from ..config import get_config
from ..utils import read_rekordbox6_asar
from ..anlz import get_anlz_paths, read_anlz_files
from . import tables

if sys.platform.lower() == "darwin":
    try:
        from pysqlcipher3 import dbapi2 as sqlite3
    except ImportError:
        import sqlite3
else:
    import sqlite3

logger = logging.getLogger(__name__)

rb6_config = get_config("rekordbox6")


def _get_masterdb_key():
    # See https://www.reddit.com/r/Rekordbox/comments/qou6nm/key_to_open_masterdb_file/

    # Read password key from app.asar file
    asar_data = read_rekordbox6_asar(rb6_config["install_dir"])
    match = re.search('pass: ".(.*?)"', asar_data).group(0)
    password = match.replace("pass: ", "").strip('"')
    password_bytes = password.encode()

    # Decode database key data
    encoded_key_data = rb6_config["dp"]  # from 'options.json'
    decoded_key_data = base64.standard_b64decode(encoded_key_data)

    # Decrypt database key
    cipher = blowfish.Cipher(password_bytes)
    decrypted_bytes = b"".join(cipher.decrypt_ecb(decoded_key_data))
    database_key = decrypted_bytes.decode()

    return database_key


def create_rekordbox_engine(path="", unlock=True, sql_driver=None, echo=None):
    if not path:
        path = rb6_config["db_path"]

    if not os.path.exists(path):
        raise FileNotFoundError(f"File '{path}' does not exist!")
    logger.info("Opening %s", path)

    # Open database
    if unlock:
        key = _get_masterdb_key()
        logger.info("Key: %s", key)
        if sql_driver is None:
            # Use default sqlite3 package
            # This requires that the 'sqlite3.dll' was replaced by
            # the 'sqlcipher.dll' (renamed to 'sqlite3.dll')
            sql_driver = sqlite3
        url = f"sqlite+pysqlcipher://:{key}@/{path}?"
        engine = create_engine(url, module=sql_driver, echo=echo)
    else:
        engine = create_engine(f"sqlite:///{path}", echo=echo)

    return engine


def open_rekordbox_database(path="", unlock=True, sql_driver=None):
    warnings.warn(
        "This method is deprecated and will be removed in a future version!"
        "Use the `Rekordbox6Database` class instead!",
        DeprecationWarning,
    )
    if not path:
        path = rb6_config["db_path"]

    if not os.path.exists(path):
        raise FileNotFoundError(f"File '{path}' does not exist!")
    logger.info("Opening %s", path)

    # Open database
    if sql_driver is None:
        # Use default sqlite3 package
        # This requires that the 'sqlite3.dll' was replaced by
        # the 'sqlcipher.dll' (renamed to 'sqlite3.dll')
        sql_driver = sqlite3
    con = sql_driver.connect(path)

    if unlock:
        # Read and decode master.db key
        key = _get_masterdb_key()
        logger.info("Key: %s", key)
        # Unlock database
        con.execute(f"PRAGMA key='{key}'")

    # Check connection
    try:
        con.execute("SELECT name FROM sqlite_master WHERE type='table';")
    except sqlite3.DatabaseError as e:
        msg = f"Opening database failed: '{e}'. Check if the database key is correct!"
        raise sqlite3.DatabaseError(msg)
    else:
        logger.info("Database unlocked!")

    return con


class Rekordbox6Database:
    def __init__(self, path="", unlock=True):
        self.engine = create_rekordbox_engine(path, unlock=unlock)
        self.Session = sessionmaker(bind=self.engine)
        self.session = self.Session()

        self._db_dir = os.path.normpath(rb6_config["db_dir"])
        self._anlz_root = os.path.join(self._db_dir, "share")

    def open(self):
        self.session = self.Session()

    def close(self):
        self.session.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def query(self, *entities, **kwargs):
        return self.session.query(*entities, **kwargs)

    # -- Djmd Tables -------------------------------------------------------------------

    def get_active_censor(self, **kwargs):
        return self.query(tables.DjmdActiveCensor).filter_by(**kwargs)

    def get_album(self, **kwargs):
        return self.query(tables.DjmdAlbum).filter_by(**kwargs)

    def get_artist(self, **kwargs):
        return self.query(tables.DjmdArtist).filter_by(**kwargs)

    def get_category(self, **kwargs):
        return self.query(tables.DjmdCategory).filter_by(**kwargs)

    def get_color(self, **kwargs):
        return self.query(tables.DjmdColor).filter_by(**kwargs)

    def get_content(self, **kwargs):
        query = self.query(tables.DjmdContent).filter_by(**kwargs)
        return query

    def search_content(self, search):

        columns = ["Title", "Commnt", "SearchStr"]

        results = list()
        for col in columns:
            obj = getattr(tables.DjmdContent, col)
            res = self.query(tables.DjmdContent).filter(obj.contains(search))
            results.extend(res.all())

        rel_columns = {
            "Album": (tables.DjmdAlbum, "Name"),
            "Artist": (tables.DjmdArtist, "Name"),
            "Composer": (tables.DjmdArtist, "Name"),
            "Remixer": (tables.DjmdArtist, "Name"),
            "Genre": (tables.DjmdGenre, "Name"),
            "Key": (tables.DjmdKey, "ScaleName"),
        }
        for col, (table, attr) in rel_columns.items():
            obj = getattr(tables.DjmdContent, col)
            item = getattr(table, attr)
            res = self.query(tables.DjmdContent).join(obj).filter(item.contains(search))
            results.extend(res.all())

        return results

    def get_cue(self, **kwargs):
        query = self.query(tables.DjmdCue).filter_by(**kwargs)
        return query

    def get_device(self, **kwargs):
        return self.query(tables.DjmdDevice).filter_by(**kwargs)

    def get_genre(self, **kwargs):
        return self.query(tables.DjmdGenre).filter_by(**kwargs)

    def get_history(self, **kwargs):
        return self.query(tables.DjmdHistory).filter_by(**kwargs)

    def get_history_songs(self, id_):
        return self.query(tables.DjmdSongHistory).filter_by(HistoryID=id_)

    def get_hot_cue_banklist(self, **kwargs):
        return self.query(tables.DjmdHotCueBanklist).filter_by(**kwargs)

    def get_hot_cue_banklist_songs(self, id_):
        return self.query(tables.DjmdSongHotCueBanklist).filter_by(HotCueBanklistID=id_)

    def get_key(self, **kwargs):
        return self.query(tables.DjmdKey).filter_by(**kwargs)

    def get_label(self, **kwargs):
        return self.query(tables.DjmdLabel).filter_by(**kwargs)

    def get_menu_items(self, **kwargs):
        return self.query(tables.DjmdMenuItems).filter_by(**kwargs)

    def get_mixer_param(self, **kwargs):
        return self.query(tables.DjmdMixerParam).filter_by(**kwargs)

    def get_my_tag(self, **kwargs):
        return self.query(tables.DjmdMyTag).filter_by(**kwargs)

    def get_my_tag_songs(self, id_):
        return self.query(tables.DjmdSongMyTag).filter_by(MyTagID=id_)

    def get_playlist(self, **kwargs):
        return self.query(tables.DjmdPlaylist).filter_by(**kwargs)

    def get_playlist_songs(self, id_):
        return self.query(tables.DjmdSongPlaylist).filter_by(PlaylistID=id_)

    def get_property(self, **kwargs):
        return self.query(tables.DjmdProperty).filter_by(**kwargs)

    def get_related_tracks(self, **kwargs):
        return self.query(tables.DjmdRelatedTracks).filter_by(**kwargs)

    def get_related_tracks_songs(self, id_):
        return self.query(tables.DjmdSongRelatedTracks).filter_by(RelatedTracksID=id_)

    def get_sampler(self, **kwargs):
        return self.query(tables.DjmdSampler).filter_by(**kwargs)

    def get_sampler_songs(self, id_):
        return self.query(tables.DjmdSongSampler).filter_by(SamplerID=id_)

    def get_tag_list_songs(self, id_):
        return self.query(tables.DjmdSongTagList).filter_by(ID=id_)

    def get_sort(self, **kwargs):
        return self.query(tables.DjmdSort).filter_by(**kwargs)

    # -- Other tables ------------------------------------------------------------------

    def get_agent_registry(self, **kwargs):
        return self.query(tables.AgentRegistry).filter_by(**kwargs)

    def get_cloud_agent_registry(self, **kwargs):
        return self.query(tables.CloudAgentRegistry).filter_by(**kwargs)

    def get_content_active_censor(self, **kwargs):
        return self.query(tables.ContentActiveCensor).filter_by(**kwargs)

    def get_content_cue(self, **kwargs):
        return self.query(tables.ContentCue).filter_by(**kwargs)

    def get_content_file(self, **kwargs):
        return self.query(tables.ContentFile).filter_by(**kwargs)

    def get_hot_cue_banklist_cue(self, **kwargs):
        return self.query(tables.HotCueBanklistCue).filter_by(**kwargs)

    def get_image_file(self, **kwargs):
        return self.query(tables.ImageFile).filter_by(**kwargs)

    def get_setting_file(self, **kwargs):
        return self.query(tables.SettingFile).filter_by(**kwargs)

    def get_uuid_map(self, **kwargs):
        return self.query(tables.UuidIDMap).filter_by(**kwargs)

    # ==================================================================================

    def get_mysetting_paths(self):
        paths = list()
        for item in self.get_setting_file():
            paths.append(os.path.join(self._db_dir, item.Path.lstrip("/\\")))
        return paths

    def get_anlz_dir(self, content_id):
        item = self.get_content(ID=content_id).one()
        dat_path = os.path.normpath(item.AnalysisDataPath).strip("\\/")
        path = os.path.join(self._anlz_root, os.path.dirname(dat_path))
        assert os.path.exists(path)
        return path

    def get_anlz_paths(self, content_id):
        root = self.get_anlz_dir(content_id)
        return get_anlz_paths(root)

    def read_anlz_files(self, content_id):
        root = self.get_anlz_dir(content_id)
        return read_anlz_files(root)
